/*
    libresize 0.08
    Copyright (C) 1998-1999  Andrew Clausen  <clausen@alphalink.com.au>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    I can also be contacted at:

    Andrew Clausen
    18 Shaw St
    Ashwood, 3147
    Victoria, Australia

    Some of this code is from mkdosfs (Dave Hudson dave@humbug.demon.co.uk)
*/

#include "libresize.h"

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>

/*
 trimfs_init_boot_sector()

 - reads in the boot sector
 - gets important info out of boot sector, and puts in main structure
 - performs sanity checks
 - returns 1 on success, 0 on failure

parameters:
  filesystem    an empty structure to fill.

*/
int
trimlib_init_boot_sector (msdos_filesystem *fs) {
    int       status, i;
    char    **ptr;

    if (!trimlib_seek_read (fs, &fs->boot_sector, 0, sizeof(msdos_boot_sector)))
        return 0;

    if (fs->boot_sector.boot_sign != 0xAA55) {
        trimlib_error (fs, "Invalid signature for a MS-based filesystem.");
        return 0;
    }

    fs->sector_size = 
        fs->boot_sector.sector_size[0] + fs->boot_sector.sector_size[1] * 256;

    fs->root_dir_entry_count = 
        fs->boot_sector.dir_entries[0] + fs->boot_sector.dir_entries[1] * 256;

    fs->sector_count = 
        fs->boot_sector.sectors[0] + fs->boot_sector.sectors[1] * 256;
    if (fs->sector_count == 0) fs->sector_count = fs->boot_sector.total_sect;

    fs->cluster_size = fs->boot_sector.cluster_size * fs->sector_size; 

    fs->fat_table_count = fs->boot_sector.fats;
    if (fs->fat_table_count != 2) {
        trimlib_error (fs, "Weird number of FATs: %d, not 2.",
                       fs->fat_table_count);
        return 0;
    }

    fs->root_dir_sector_count
        = fs->root_dir_entry_count * sizeof (directory_entry) / fs->sector_size;

    fs->boot_sector_offset = 0;
    fs->info_sector_offset = fs->boot_sector.fat32_info_sector *fs->sector_size;

    fs->fat_offset = fs->boot_sector_offset
                        + fs->boot_sector.reserved * fs->sector_size;

    if (fs->boot_sector.fat_length == 0) {
        // It's definitely FAT32
        fs->fs_type = FAT32;
        fs->root_offset = fs->fat_offset
                            + fs->boot_sector.fat32_fat_length * fs->sector_size
                                * fs->fat_table_count;

        fs->cluster_offset = fs->root_offset - 2 * fs->cluster_size;

        fs->root_cluster = fs->boot_sector.fat32_root_dir_cluster;

        fs->fat_size = fs->boot_sector.fat32_fat_length * fs->sector_size;
        fs->fat_entry_count = fs->fat_size / 4;
    } else {
        fs->fs_type = 0;  /* unknown yet - it isn't FAT32 though! */
        fs->root_offset = fs->fat_offset
                            + fs->boot_sector.fat_length * fs->sector_size
                                * fs->fat_table_count;
        fs->cluster_offset
            = fs->root_offset
                 + fs->root_dir_entry_count * sizeof (directory_entry)
                 - 2 * fs->cluster_size;

        fs->fat_size = fs->boot_sector.fat_length * fs->sector_size;
        fs->fat_entry_count = fs->fat_size / 2;
    }

    /* - 2 because clusters 0 & 1 doesn't exist */
    fs->cluster_count = 
        (fs->sector_count * fs->sector_size - fs->cluster_offset)
            / fs->cluster_size - 2;

    fs->dir_entries_per_cluster = fs->cluster_size / sizeof (directory_entry);

/*
  // FIXME: put this back in
    for (ptr = &fs_types[0], i=0; *ptr != (char *)NULL; ptr++, i++) {
        if (strncmp(*ptr, fs->boot_sector.fs_type, 5) == 0) {
             fs->fs_type = i;
             break;
        }
    }

    if (*ptr == NULL) {
        if (strcmp(fs->boot_sector.fs_type, "") == 0)
            trimlib_error (fs, "unknown filesystem type");
        else
            trimlib_error (fs, "unknown filesystem type `%.8s'",
                           fs->boot_sector.fs_type);

        return 0;
    }
*/

/* FIXME: find out what these numbers REALLY are */
    if (fs->cluster_count < 0xff4) fs->fs_type = FAT12;
    if (fs->cluster_count >= 0xff4 && fs->cluster_count < 0xfff4)
        fs->fs_type = FAT16;
    if (fs->cluster_count >= 0xfff4) fs->fs_type = FAT32;

    if (fs->fs_type == FAT12) {
        trimlib_error (fs, "FAT12 filesystem not supported.");
        return 0;
    }

    if (fs->sector_count < 32) {
        trimlib_error (fs, "Too few sectors for viable file system.");
        return 0;
    }

    return 1;
}

int
trimlib_write_boot_sector (msdos_filesystem *fs) {
    int status;

    if (!trimlib_seek_write (fs, (char *) &fs->boot_sector,
                             fs->boot_sector_offset, 512)) {
        trimlib_error (fs, "error while writing the boot sector");
        return 0;
    }

    if (fs->fs_type == FAT32) {
        // write backup
        if (!trimlib_seek_write (fs, (char *) &fs->boot_sector,
                  fs->boot_sector.fat32_backup_sector * fs->sector_size, 512)) {
            trimlib_error (fs, "error while writing the backup boot sector");
            return 0;
        }
    }

    return 1;
}

int
trimlib_read_info_sector (msdos_filesystem *fs) {
    if (!trimlib_seek_read (fs, &fs->info_sector, fs->info_sector_offset,
                            sizeof(msdos_info_sector)))
        return 0;

    if (fs->info_sector.signature != 0x61417272) {
        trimlib_error (fs, "Invalid information sector signature 0x%x",
                       fs->info_sector.signature);
        return 0;
    }
    return 1;
}

int
trimlib_write_info_sector (msdos_filesystem *fs) {
    fs->info_sector.signature = 0x61417272;
    fs->info_sector.free_clusters = fs->free_cluster_count;
    fs->info_sector.next_cluster = 2;

    if (!trimlib_seek_write (fs, &fs->info_sector, fs->info_sector_offset,
                            sizeof(msdos_info_sector)))
        return 0;

    return 1;
}
