/*
    libresize 0.08
    Copyright (C) 1998-1999  Andrew Clausen  <clausen@alphalink.com.au>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    I can also be contacted at:

    Andrew Clausen
    18 Shaw St
    Ashwood, 3147
    Victoria, Australia

*/

#include "libresize.h"
#include "traverse.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*
    prints out the sequence of clusters for a given file chain, beginning
    at start_cluster.
*/
static void
print_chain (msdos_filesystem *fs, cluster_t start_cluster) {
    cluster_t   this_cluster;
    int         this_row;

    this_row = 0;
    for (this_cluster = start_cluster;
         !trimlib_is_eof (fs, this_cluster);
         this_cluster = trimlib_get_fat (fs, this_cluster)) {
        printf ("  %ld", this_cluster);

        if (++this_row == 7) {
            printf ("\n");
            this_row = 0;
        }
    }
    printf ("\n");
}

/*
    traverse the FAT for a file/directory, marking each entry's flag
    to "flag".
*/
static int
trimlib_flag_traverse_fat (msdos_filesystem *fs, char *chain_name,
                           long cluster, int flag) {
    long    this_cluster;
    
    for (this_cluster = cluster;
         !trimlib_is_eof (fs, this_cluster);
         this_cluster = trimlib_get_fat (fs, this_cluster)) {

        if (this_cluster == 0) {
            trimlib_error (fs, "Bad FAT: unterminated chain for %s",
                           chain_name);
            return 0;
        }

        if (fs->fat_flag_map[this_cluster] != 0) {
            trimlib_error (fs, "Bad FAT: cluster %d is cross-linked for %s",
                           (int) this_cluster, chain_name);
            return 0;
        }

        if (flag == FAT_FLAGS_DIRECTORY) fs->total_dir_clusters++;
        fs->fat_flag_map[this_cluster] = flag;
    }

    return 1;
}

/*
    recursively traverses a directory, flagging all clusters in the process.
    It frees the traverse_info structure before returning.
*/
static int
trimlib_flag_traverse_dir (trimlib_traverse_info *traverse_info) {
    directory_entry        *this_entry;
    trimlib_traverse_info  *subdir_traverse_info;
    char                    file_name[512];
    char                   *file_name_start;
    cluster_t               first_cluster;

    strcpy (file_name, traverse_info->dir_name);
    file_name_start = file_name + strlen(file_name);
    
    while ( (this_entry = trimlib_get_next (traverse_info)) ) {
        if (this_entry->name[0] == '.') continue;   /* skip . and .. entries */
        if (this_entry->name[0] == DELETED_FLAG) continue; /* skip del'd files*/
        if (this_entry->name[0] == 0) continue;	    /* skip empty slots */
        if (this_entry->name[0] == 0xF6) continue;  /* it's in the kernel ;-) */

        trimlib_file_get_name (this_entry, file_name_start);
        first_cluster = trimlib_file_get_first_cluster (traverse_info->fs,
                                                        this_entry);
        
        if (trimlib_is_directory (this_entry)) {
            if (!trimlib_flag_traverse_fat (traverse_info->fs, file_name,
                                           first_cluster, FAT_FLAGS_DIRECTORY)){
#ifndef USE_NEWT
                printf ("%s chain:\n", file_name);
                print_chain (traverse_info->fs, first_cluster);
#endif
                return 0;
            }

            subdir_traverse_info
                = trimlib_traverse_directory (traverse_info, this_entry);
            if (!trimlib_flag_traverse_dir (subdir_traverse_info)) return 0;
        } else if (trimlib_is_file (this_entry)) {
            if (this_entry->length == 0) continue;      /* skip empty files */

            if (!trimlib_flag_traverse_fat (traverse_info->fs, file_name,
                                            first_cluster, FAT_FLAGS_FILE)) {
#ifndef USE_NEWT
                printf ("%s chain:\n", file_name);
                print_chain (traverse_info->fs, first_cluster);
#endif
                return 0;
            }
#ifdef DEBUG
            printf ("%s\n", file_name);
#endif
        }
    }

    trimlib_traversal_complete (traverse_info);
    return 1;
}

/*  
    fills in fs->fat_flag_map.  Each FAT entry is flagged as either
    FAT_FLAGS_FREE, FAT_FLAGS_FILE or FAT_FLAGS_DIRECTORY.
*/
int
trimlib_flag_clusters (msdos_filesystem *fs) {
    trimlib_traverse_info  *traverse_info;
    
    /* set all clusters to unused as a default */
    memset (fs->fat_flag_map, FAT_FLAGS_FREE, fs->fat_entry_count);
    fs->total_dir_clusters = 0;

    if (fs->fs_type == FAT32) {
        traverse_info = trimlib_traversal_begin (fs, fs->root_cluster, "\\");
        if (!trimlib_flag_traverse_dir (traverse_info)) return 0;
        if (!trimlib_flag_traverse_fat (fs, "the root directory",
                                        fs->root_cluster,
                                        FAT_FLAGS_DIRECTORY)) {
#ifndef USE_NEWT
                printf ("root directory chain:\n");
                print_chain (traverse_info->fs, fs->root_cluster);
#endif
            return 0;
        }
    } else {
        traverse_info = trimlib_traversal_begin (fs, FAT_ROOT, "\\");
        if (!trimlib_flag_traverse_dir (traverse_info)) return 0;
    }

    return 1;
}

