/*
    libresize 0.08
    Copyright (C) 1998-1999  Andrew Clausen  <clausen@alphalink.com.au>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    I can also be contacted at:

    Andrew Clausen
    18 Shaw St
    Ashwood, 3147
    Victoria, Australia

    Some of this code is from mkdosfs (Dave Hudson dave@humbug.demon.co.uk)
*/

#ifndef TRIMLIB_H_INCLUDED
#define TRIMLIB_H_INCLUDED

#include "../libfdisk/libfdisk.h"

//#define DEBUG
//#define READ_ONLY

#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#define MAX_BUFFER_SIZE 4 * 1024 * 1024  /* buffer size for each fs in bytes */

typedef unsigned int                cluster_t;

typedef struct _msdos_boot_sector   msdos_boot_sector;
typedef struct _msdos_filesystem    msdos_filesystem;
typedef struct _directory_entry     directory_entry;
typedef struct _msdos_info_sector   msdos_info_sector;

struct _msdos_boot_sector {
    unsigned char boot_jump[3];   /* 00: Boot strap short or near jump */
    char system_id[8];            /* 03: Name - can be used to special case
                                   partition manager volumes */
    unsigned char sector_size[2]; /* 0b: bytes per logical sector */
    unsigned char cluster_size;   /* 0d: sectors/cluster */
    unsigned short reserved;      /* 0e: reserved sectors */
    unsigned char fats;           /* 10: number of FATs */
    unsigned char dir_entries[2]; /* 11: number of root directory entries */
    unsigned char sectors[2];     /* number of sectors: total_sect supersedes */
    unsigned char media;          /* 15: media code - unused */
    unsigned short fat_length;    /* 16: sectors/FAT for FAT12/16 */
    unsigned short secs_track;    /* 18: sectors per track */
    unsigned short heads;         /* 1a: number of heads */
    unsigned int hidden;          /* 1c: hidden sectors (unused) */
    unsigned int total_sect;      /* 20: number of sectors (if sectors == 0) */

// FAT32-only entries
    unsigned int fat32_fat_length;        /* 24: size of FAT in sectors */
    unsigned short fat32_flags;           /* 28: bit8: fat mirroring,
                                                 low4: active fat */
    unsigned short fat32_version;         /* 2a: minor * 256 + major */
    unsigned int fat32_root_dir_cluster;  /* 2c: */
    unsigned short fat32_info_sector;     /* 30: */
    unsigned short fat32_backup_sector;   /* 32: */

// Common again...
    unsigned char boot_code[458]; /* 34: Boot code (or message) */
    unsigned short boot_sign;     /* 0xAA55 */
};

struct _directory_entry {
    unsigned char       name[8];
    unsigned char       extension[3];
    unsigned char       attributes;
    unsigned char       is_upper_case_name;
    unsigned char       creation_time_low;      /* milliseconds */
    unsigned short      creation_time_high;
    unsigned short      creation_date;
    unsigned short      access_date;
    unsigned short      first_cluster_high;     /* for FAT32 */
    unsigned short      time;
    unsigned short      date;
    unsigned short      first_cluster;
    unsigned int        length;
};

struct _msdos_info_sector {
    char            unused [0x1e4];
    unsigned int    signature;      /* should be 0x61417272 */
    unsigned int    free_clusters;  /* -1 for unknown */
    unsigned int    next_cluster;   /* most recently allocated cluster */
    char            unused2 [0x10];
};

#define FAT_FLAGS_FREE      0
#define FAT_FLAGS_FILE      1
#define FAT_FLAGS_DIRECTORY 2

struct _msdos_filesystem {
    char               *device_name;    /* name of device (eg /dev/hda) */
    HardDrive          *device;         /* libfdisk structure */
    int                 partition;      /* partition number on device */

    char               *fs_name;        /* file of fs (eg /dev/hda3) */
    int                 fs_handle;      /* handle for it */

    msdos_boot_sector   boot_sector;    /* structure of boot sector */
    msdos_info_sector   info_sector;    /* fat32-only information sector */

    long long           partition_start;
    long long           partition_size;

    int                 sector_size;
    int                 sector_count;
    int                 cluster_size;
    cluster_t           cluster_count;
    cluster_t           fat_entry_count;        /* no of FAT entries */
    int                 fat_table_count;
    int                 fat_size;               /* single FAT in bytes */
    int                 fs_type;                /* FAT12, FAT16, FAT32 */
    int                 dir_entries_per_cluster;

    long                boot_sector_offset;     /* offsets are in bytes */
    long                info_sector_offset;     // FAT32 only
    long                fat_offset;
    long                root_offset;            // non-FAT32
    long                cluster_offset;

    cluster_t           used_cluster_count;
    cluster_t           free_cluster_count;
    cluster_t           bad_cluster_count;
    cluster_t           last_free_cluster;
    cluster_t           root_cluster;           // FAT32 only
    long                root_dir_entry_count;   // non-FAT32
    long                root_dir_sector_count;  // non-FAT32
    cluster_t           total_dir_clusters;

    char               *buffer;
    char               *buffer2;
    cluster_t           buffer_size;            /* buffer size in clusters */

    void               *fats[2];
    void               *new_fat;
    long               *fat_remap;
    char               *fat_flag_map;

    char                error_msg [256];
    char                scratch [512];
};

#define FAT12 0
#define FAT16 1
#define FAT32 2

#define FAT_ROOT          0x0000

#define DELETED_FLAG      0xe5
#define VOLUME_LABEL_ATTR 0x08
#define VFAT_ATTR         0x0f
#define DIRECTORY_ATTR    0x10

#define MAX_FAT12_CLUSTERS  4086      
#define MAX_FAT16_CLUSTERS  65526

extern char *fs_types[];

extern int trimlib_init_filesystem (msdos_filesystem *filesystem,
                                    char *device_name, int partition);

extern int trimlib_prepare_filesystem (msdos_filesystem *fs);

extern void trimlib_unprepare_system (msdos_filesystem *fs);

extern void trimlib_print_info (msdos_filesystem *filesystem);

extern int trimlib_resize_filesystem (msdos_filesystem *filesystem,
                                      long long new_size);

extern int trimlib_init_boot_sector (msdos_filesystem *filesystem);

#include "backup.h"

#endif /* TRIMLIB_H_INCLUDED */
