/*
    libresize 0.08
    Copyright (C) 1998-1999  Andrew Clausen  <clausen@alphalink.com.au>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    I can also be contacted at:

    Andrew Clausen
    18 Shaw St
    Ashwood, 3147
    Victoria, Australia

*/

#include "libresize.h"
#include "traverse.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define NO_CLUSTER -1

cluster_t
trimlib_file_get_first_cluster (msdos_filesystem *fs,
                                directory_entry *dir_entry) {
    switch (fs->fs_type) {
        case FAT16:
            return dir_entry->first_cluster;
            
        case FAT32:
            return dir_entry->first_cluster_high * 65536L
                      + dir_entry->first_cluster;
    }
}

void
trimlib_file_set_first_cluster (msdos_filesystem *fs,
                                directory_entry *dir_entry, cluster_t cluster) {
    switch (fs->fs_type) {
        case FAT16:
            dir_entry->first_cluster = cluster;
            break;
            
        case FAT32:
            dir_entry->first_cluster = cluster & 0xffff;
            dir_entry->first_cluster_high = cluster / 0x10000;
            break;
    }
}

int
entries_per_buffer (trimlib_traverse_info *traverse_info) {
    if (traverse_info->is_legacy_root_dir)
        return traverse_info->fs->sector_size / sizeof (directory_entry);
    else
        return traverse_info->fs->cluster_size / sizeof (directory_entry);
}

trimlib_traverse_info *
trimlib_traversal_begin (msdos_filesystem *fs, int start_cluster,
                         char *dir_name) {
    trimlib_traverse_info  *traverse_info;

    traverse_info
        = (trimlib_traverse_info *) malloc (sizeof (trimlib_traverse_info));
    if (traverse_info == NULL) {
        trimlib_error (fs, "malloc() failed for traversal structure");
        return NULL;
    }

    traverse_info->dir_entries = (directory_entry *) malloc (fs->cluster_size);
    if (traverse_info->dir_entries == NULL) {
        trimlib_error (fs, "malloc() failed for traversal buffer");
        return NULL;
    }

    strcpy (traverse_info->dir_name, dir_name);
    traverse_info->fs = fs;
    traverse_info->start_cluster = start_cluster;
    traverse_info->this_cluster = NO_CLUSTER;
    traverse_info->is_legacy_root_dir
        = (fs->fs_type == FAT16) && (start_cluster == 0);

    return traverse_info;
}

void
trimlib_traversal_complete (trimlib_traverse_info *traverse_info) {
    free (traverse_info->dir_entries);
    free (traverse_info);
}

trimlib_traverse_info *
trimlib_traverse_directory (trimlib_traverse_info *traverse_info,
                            directory_entry *parent) {
    trimlib_traverse_info  *child;
    
    child = trimlib_traversal_begin (
        traverse_info->fs,
        trimlib_file_get_first_cluster (traverse_info->fs, parent), "");

    strcpy (child->dir_name, traverse_info->dir_name);
    trimlib_file_get_name (parent, child->dir_name + strlen(child->dir_name));
    strcat (child->dir_name, "\\");

    return child;
}

/*
    returns 1 if there are no more directory entries in the directory being
    traversed, 0 otherwise.
*/
static int
is_end_of_dir (trimlib_traverse_info *traverse_info) {
    int                 i;
    msdos_filesystem   *fs = traverse_info->fs;
    
    /* have we even started reading in the dir? */
    if (traverse_info->this_cluster == NO_CLUSTER) return 0;

    /* is there another cluster/root dir entry? */
    if (traverse_info->this_dir_entry_index
           >= entries_per_buffer (traverse_info)) {
        if (traverse_info->is_legacy_root_dir) {
            return traverse_info->this_cluster >= fs->root_dir_sector_count;
        } else {
            return trimlib_is_eof (fs, traverse_info->this_cluster);
        }
    }

    /* a NULL entry terminates */
    for (i=0; i<sizeof(directory_entry); i++)
        if ( ((char *) traverse_info->this_dir_entry)[i] != 0) return 0;

    return 1;
}

directory_entry *
trimlib_get_next (trimlib_traverse_info *traverse_info) {
    msdos_filesystem   *fs = traverse_info->fs;
    long                read_offset;
    
    if (traverse_info->this_cluster != NO_CLUSTER) {
        traverse_info->this_dir_entry++;
        traverse_info->this_dir_entry_index++;

	if (is_end_of_dir (traverse_info)) return NULL;
    }

    if (traverse_info->is_legacy_root_dir) {
        if (traverse_info->this_dir_entry_index
                >= entries_per_buffer (traverse_info)
            || traverse_info->this_cluster == NO_CLUSTER) {
            
            if (traverse_info->this_cluster == NO_CLUSTER)
                traverse_info->this_cluster = 0;
            else
                traverse_info->this_cluster++;

            read_offset = fs->root_offset
                           + traverse_info->this_cluster * fs->sector_size;
            
            if (!trimlib_seek_read (fs, (void *) traverse_info->dir_entries,
                                    read_offset, fs->sector_size))
                return NULL;

            traverse_info->this_dir_entry = traverse_info->dir_entries;
            traverse_info->this_dir_entry_index = 0;
        }
    } else {
        if (traverse_info->this_dir_entry_index
                >= entries_per_buffer (traverse_info)
            || traverse_info->this_cluster == NO_CLUSTER) {
            
            if (traverse_info->this_cluster == NO_CLUSTER)
                traverse_info->this_cluster = traverse_info->start_cluster;
            else
                traverse_info->this_cluster
                     = trimlib_get_fat (fs, traverse_info->this_cluster);
            
            if (!trimlib_read_cluster (fs, (void *) traverse_info->dir_entries,
                                       traverse_info->this_cluster))
                return NULL;

            traverse_info->this_dir_entry = traverse_info->dir_entries;
            traverse_info->this_dir_entry_index = 0;
        }
    }
    return traverse_info->this_dir_entry;
}

int
trimlib_is_file (directory_entry *dir_entry) {
    if (dir_entry->attributes == VFAT_ATTR) return 0;
    if (dir_entry->attributes & DIRECTORY_ATTR) return 0;
    if (dir_entry->attributes & VOLUME_LABEL_ATTR) return 0;
    if (dir_entry->name[0] == DELETED_FLAG) return 0;
    if (dir_entry->name[0] == 0) return 0;
    if (dir_entry->name[0] == 0xF6) return 0;
    return 1;
}

int
trimlib_is_directory (directory_entry *dir_entry) {
    if (dir_entry->name[0] == DELETED_FLAG) return 0;
    if (dir_entry->name[0] == 0) return 0;
    if (dir_entry->name[0] == 0xF6) return 0;
    if (dir_entry->attributes == VFAT_ATTR) return 0;
    return (dir_entry->attributes & DIRECTORY_ATTR) > 0;
}

int
trimlib_has_first_cluster (msdos_filesystem *fs, directory_entry *dir_entry) {
    cluster_t   first_cluster;
    
    if (!trimlib_is_file (dir_entry) && !trimlib_is_directory (dir_entry))
        return 0;

    first_cluster = trimlib_file_get_first_cluster (fs, dir_entry);
    if (first_cluster == 0 || trimlib_is_eof (fs, first_cluster)) return 0;

    return 1;
}

/*
    decrypts silly DOS names to FILENAME.EXT
*/
void
trimlib_file_get_name (directory_entry *dir_entry, char *result) {
    int     i;
    char   *src;

    src = dir_entry->name;

    for (i=0; i<8; i++) {
        if (src[i] == ' ' || src[i] == 0) break;
        *result++ = src[i];
    }

    if (src[8] != ' ' && src[8] != 0) {
        *result++ = '.';
        for (i=8; i<11; i++) {
            if (src[i] == ' ' || src[i] == 0) break;
            *result++ = src[i];
        }
    }

    *result = 0;
}

